/**
 * Dagger build and deploy functions
 */
import {
  dag,
  Container,
  Directory,
  File,
  object,
  func,
} from "@dagger.io/dagger";

@object()
class HelloDagger {
  /**
   * Return the result of running unit tests
   */
  @func()
  build(source: Directory): Container {
    const executable = this.buildEnv(source)
      .withExec(["go", "build"])
      .file("./helloworld");

    return dag
      .container()
      .from("ubuntu:latest")
      .withFile("/usr/bin/helloworld", executable)
      .withExposedPort(9000)
      .withEntrypoint(["helloworld"]);
  }

  /**
   * Return the result of running unit tests
   */
  @func()
  async test(source: Directory): Promise<string> {
    // TODO return error code?
    return this.buildEnv(source).withExec(["go", "test"]).stdout();
  }

  /**
   * Run trivy on the source code
   */
  @func()
  async securityScan(source: Directory): Promise<string> {
    return dag.trivy().fs(source).contents();
  }

  /**
   * Run golandci-lint on the project
   */
  @func()
  lint(source: Directory): Container {
    return dag
      .container()
      .from("golangci/golangci-lint:v1.61.0")
      .withDirectory("/src", source)
      .withWorkdir("/src")
      .withExec(["golangci-lint", "run", "-v"]);
  }

  @func()
  async publish(source: Directory, target?: string): Promise<string> {
    // publish can only push to external registries
    // to publish to your local docker, you have to export the tarball and use `docker load`

    // This example is somewhat useless as one could just chain `publish` to `call build`

    // publish automatically picks up the local auth from the host
    // otherwise, use `withRegistryAuth()`
    return this.build(source).publish(
      target ?? "mtr.devops.telekom.de/patrick_sy/hello-dagger:latest",
    );
  }

  /**
   * Build a ready-to-use development environment
   */
  @func()
  buildEnv(source: Directory): Container {
    return dag
      .container()
      .from("golang:1.23")
      .withDirectory("/src", source)
      .withWorkdir("/src");
  }
}
